/******************************************************************
* Phoenix Lexical Analyzer Generator.                             *
* From intermediate structures to DFA.                            *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"
#include "ph_error.h"

#define PUSH    {if(PushStack()==FALSE)return FALSE; if(LexTrace)if(TraceStr("PUSH\r\n")==FALSE)return FALSE;}
#define POP     {PoolFreeList(lpStack[StackIndex].Firstpos);PoolFreeList(lpStack[StackIndex].Lastpos);\
                 StackIndex--;if(LexTrace)if(TraceStr("\r\nPOP\r\n")==FALSE)return FALSE;}

static WORD iRegExp;                // Index of current regular expression
static HPLEXP2REGEXP hpRegExp;      // Huge pointer to current regular expression structure:
                                    // hpRegExp = hpP2RegExp + iRegExp
static LPWORD lpRg;                 // The current regular expression: hpP2RegExp[iRegExp].RegExp
static WORD LookAheadIndex;         // Index of next symbol in current regular expression
static WORD LookAhead;              // Current symbol in current regular expression:
                                    // next value of LookAhead = lpRg[LookAheadIndex]

static WORD suivant;
static BOOL str;


/* Add a position to a linked list in the pool. The list grows backwards, and like
   a LIFO stack: the new position becomes the new first position of the list.
   the new head of list is returned via NewHead. */
BOOL PoolAddPosToList(DWORD List, long int Position, LPDWORD NewHead)
{
    if(Holes == 0)
    {
        /* Increment the top of the pool */
        if(lpPool == NULL || nPool + 1 >= nMaxPool - 1)
        {
            if(PoolPush(Position) == FALSE) return FALSE;
        }
        else
            lpPool[++nPool].Pos = Position;
        *NewHead = nPool;
    }
    else
    {
        /* Use the first available hole */
        lpPool[Holes].Pos = Position;
        *NewHead = Holes;
        Holes = lpPool[Holes].Next;
    }

    lpPool[*NewHead].Next = List;

    return TRUE;
}

/* Free a linked list of positions: add it to the list of holes */
__inline void PoolFreeList(DWORD List)
{
    DWORD EndList;

    if(List == 0) return;

    /* Traverse the list to find its end */
    EndList = List;
    while(lpPool[EndList].Next) EndList = lpPool[EndList].Next;

    /* Connect end of list to list of holes */
    lpPool[EndList].Next = Holes;
    Holes = List;
}

/* Compute the length of a linked list */
__inline DWORD PoolListLength(DWORD List)
{
    DWORD Length;

    Length = 0;
    while(List) {List = lpPool[List].Next; Length++;}

    return Length;
}

/* Make sure that the next character matches the character desired by the syntax rules, then
   move LookAhead ahead */
BOOL Match(WORD c)
{
    if(LookAhead != c)
    {
        LexErrorCode = LEXERR_REG_EXP_SYNTAX;
        return FALSE;
    }
    if(GetSourceToken() == FALSE) return FALSE;

    return TRUE;
}

/* Trace: print a string */
BOOL TraceStr(LPSTR s)
{
    WORD i;
    HFILE hf;
    OFSTRUCT ofs;

    for(i = 0; i < MAX_ERROR_PARAM && LexTraceFile[i]; i++) LexErrorStringParam[i] = LexTraceFile[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(LexTraceFile, &ofs, OF_WRITE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_OPEN_FILE;
        return FALSE;
    }

    if(_llseek(hf,0L,2) == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_EOF_FILE;
        return FALSE;
    }

    if(_lwrite(hf,s,lstrlen(s))== HFILE_ERROR)
    {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

    _lclose(hf);

    return TRUE;
}

/* Trace: print Firstpos and Lastpos sets in stack */
BOOL TracePos()
{
    DWORD i,j;
    HFILE hf;
    OFSTRUCT ofs;

    for(i = 0; i < MAX_ERROR_PARAM && LexTraceFile[i]; i++) LexErrorStringParam[i] = LexTraceFile[i];   // Copy FileName to parameter
    LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM

    hf = OpenFile(LexTraceFile, &ofs, OF_WRITE);
    if (hf == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_OPEN_FILE;
        return FALSE;
    }

    if(_llseek(hf,0L,2) == HFILE_ERROR)
    {
        LexErrorCode = LEXERR_CANT_EOF_FILE;
        return FALSE;
    }

    for(j = 1; j <= StackIndex; j++)        // Index 0 of stack is unused
    {
        if(_lwrite(hf,"\r\n",2)== HFILE_ERROR)
        {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

        wsprintf((LPSTR)OutBuffer,"\t\tStackIndex=%lu\r\nNullable=%u\r\nFirstpos={",j,lpStack[j].Nullable);
        if(_lwrite(hf,(LPSTR)OutBuffer,lstrlen((LPSTR)OutBuffer))== HFILE_ERROR)
        {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

        i = lpStack[j].Firstpos;
        while(i)
        {
            if(i == lpStack[j].Firstpos)
                wsprintf((LPSTR)OutBuffer,"%ld",lpPool[i].Pos);
            else
                wsprintf((LPSTR)OutBuffer,",%ld",lpPool[i].Pos);

            if(_lwrite(hf,(LPSTR)OutBuffer,lstrlen((LPSTR)OutBuffer))== HFILE_ERROR)
            {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

            i = lpPool[i].Next;
        }
        if(_lwrite(hf,"}\r\nLastpos={",12)== HFILE_ERROR)
        {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

        i = lpStack[j].Lastpos;
        while(i)
        {
            if(i == lpStack[j].Lastpos)
                wsprintf((LPSTR)OutBuffer,"%ld",lpPool[i].Pos);
            else
                wsprintf((LPSTR)OutBuffer,",%ld",lpPool[i].Pos);

            if(_lwrite(hf,(LPSTR)OutBuffer,lstrlen((LPSTR)OutBuffer))== HFILE_ERROR)
            {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

            i = lpPool[i].Next;
        }
        if(_lwrite(hf,"}",1)== HFILE_ERROR)
        {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}
    }
    if(_lwrite(hf,"\r\n",2)== HFILE_ERROR)
    {_lclose(hf); LexErrorCode = LEXERR_CANT_WRITE_FILE; return FALSE;}

    _lclose(hf);

    return TRUE;
}

BOOL ConstructDfa()
{
    if(DfaInit() == FALSE) return FALSE;

    /* Loop on each regular expression, and build a distinct DFA every time */
    for(iRegExp = 0; iRegExp < nP2RegExp; iRegExp++)
    {
        hpRegExp = hpP2RegExp + iRegExp;
        LookAheadIndex = 0;

        lpRg = (LPWORD)hpP2RegExp[iRegExp].RegExp;
        if(GetSourceToken() == FALSE) return FALSE;
        if(LookAhead == STAR || LookAhead == OR || LookAhead == QMARK || LookAhead == PLUS)
        {LexErrorCode = LEXERR_INVALID_FIRST_CHARACTER; return FALSE;}

        if(LexTrace)
        {
            unsigned int i;
            LPSTR pCat;

            wsprintf((LPSTR)OutBuffer,"\r\n\t\t\tRegular Expression %u: ",iRegExp);
            pCat = OutBuffer + lstrlen(OutBuffer);
            for(i = 0; i < hpRegExp -> RegExpSize; i++)
            	*pCat++ = (char)(hpRegExp -> RegExp[i]);
            *pCat++ = '\r';
            *pCat++ = '\n';
            *pCat = '\0';
            if(TraceStr((LPSTR)OutBuffer) == FALSE) return FALSE;
        }

        PUSH;

        if(Regular()    == FALSE) return FALSE;
        if(Tables()     == FALSE) return FALSE;
        if(AddToken()   == FALSE) return FALSE;
        if(Clear()      == FALSE) return FALSE;
    }

    if(AttachDFAs()     == FALSE) return FALSE;

    /* Minimizing the global DFA */
    CreateInitialPartition();
    if(MinimizeDFA()    == FALSE) return FALSE;

    /* Replacing the classes by their equivalent character set */
    AddClassTransitions();

    return TRUE;
}

/* Initialization called once before global DFA construction */
BOOL DFAInit()
{
    register WORD i;
    HFILE hf;
    OFSTRUCT ofs;

    Holes = 0;
    TotalStates = 0;
    nToken = 0;
    str = FALSE;
    suivant = 0;
    lpStates = NULL;
    hpGNextState = NULL;
    hpFNextState = NULL;
    lpStack = NULL;
    lpPos = NULL;
    lpPool = NULL;

    for(i = 0; i < MAX_CHARS + MAX_CLASSES; i++) Alphabet[i] = 0;

    if(LexTrace)
    {
        /* Create trace file */
        hf = OpenFile((LPSTR)LexTraceFile, &ofs, OF_CREATE);
        if (hf == HFILE_ERROR)
        {
            LexErrorCode = LEXERR_CANT_CREATE_FILE;
            for(i = 0; i < MAX_ERROR_PARAM && LexTraceFile[i]; i++) LexErrorStringParam[i] = LexTraceFile[i];   // Copy FileName to parameter
            LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM
            return FALSE;
        }
        _lclose(hf);
    }

    return TRUE;
}

/* After DFA has been written to output, structures are to be freed */
void DFAFree()
{
    if(lpStates != NULL)
    {
        GlobalFreePtr(lpStates);
        lpStates = NULL;
    }

    if(hpGNextState != NULL)
    {
        GlobalFreePtr(hpGNextState);
        hpGNextState = NULL;
    }

    if(hpFNextState != NULL)
    {
        GlobalFreePtr(hpFNextState);
        hpFNextState = NULL;
    }

    if(lpPool != NULL)
    {
        GlobalFreePtr(lpPool);
        lpPool = NULL;
    }

    if(lpStack != NULL)
    {
        GlobalFreePtr(lpStack);
        lpStack = NULL;
    }

    if(lpPos != NULL)
    {
        GlobalFreePtr(lpPos);
        lpPos = NULL;
    }
}

BOOL Regular()
{
    BOOL Term();

    if(Term() == FALSE) return FALSE;
    while(1)
    {
        if(LookAhead == OR)
        {
            PUSH;
            if(Match(OR) == FALSE) return FALSE;
            if(Term()    == FALSE) return FALSE;
            if(Node(OR)  == FALSE) return FALSE;
        }
        else break;
    }

    return TRUE;
}

BOOL Term()
{
    BOOL Factor();

    if(Factor() == FALSE) return FALSE;
    while(1)
    {
        if(LookAhead && LookAhead <= MAX_CHARS || LookAhead >= CLASS_CODES || LookAhead == LEFT)
        {
            PUSH;
            if(Factor() == FALSE) return FALSE;
            if(Node(CAT) == FALSE) return FALSE;
        }
        else break;
    }

    return TRUE;
}

BOOL Factor()
{
    WORD c;

    if(LookAhead == LEFT)
    {
        if(Match(LEFT) == FALSE) return FALSE;
        if(Regular() == FALSE) return FALSE;
        if(Match(RIGHT) == FALSE) return FALSE;
    }
    else
    {
        if(Node(LookAhead) == FALSE) return FALSE;
        if(Match(LookAhead) == FALSE) return FALSE;
    }

    while(1)
    {
        c = LookAhead;
        if(c == STAR || c == PLUS || c == QMARK)
        {
            if(Match(c) == FALSE) return FALSE;
            if(Node(c) == FALSE) return FALSE;  // c used instead of LookAhead, because Match() changes the value of LookAhead
        }
        else break;
    }

    return TRUE;
}

/* Process a node of the syntax tree representing the current regular expression */
BOOL Node(WORD c)
{
    register DWORD i,j,p,m;
    BOOL Same;
    static DWORD NewHead;

    switch(c)
    {
        case QMARK:
            if(LexTrace) if(TraceStr("? ") == FALSE) return FALSE;
            lpStack[StackIndex].Nullable = TRUE;
            break;

        case STAR:
            if(LexTrace) if(TraceStr("* ") == FALSE) return FALSE;
            lpStack[StackIndex].Nullable = TRUE;

            i = lpStack[StackIndex].Lastpos;
            while(i)
            {
                m = lpPool[i].Pos;
                j = lpStack[StackIndex].Firstpos;
                while(j)
                {
                    Same = FALSE;
                    p = lpPos[m].Followpos;
                    while(p)
                    {
                        if(lpPool[p].Pos == lpPool[j].Pos) Same = TRUE;
                        p = lpPool[p].Next;
                    }

                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(lpPos[m].Followpos, lpPool[j].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        lpPos[m].Followpos = NewHead;
                    }

                    j = lpPool[j].Next;
                }

                i = lpPool[i].Next;
            }
            break;

        case PLUS:
            if(LexTrace) if(TraceStr("+ ") == FALSE) return FALSE;
            i = lpStack[StackIndex].Lastpos;
            while(i)
            {
                m = lpPool[i].Pos;
                j = lpStack[StackIndex].Firstpos;
                while(j)
                {
                    Same = FALSE;
                    p = lpPos[m].Followpos;
                    while(p)
                    {
                        if(lpPool[p].Pos == lpPool[j].Pos) Same = TRUE;
                        p = lpPool[p].Next;
                    }

                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(lpPos[m].Followpos, lpPool[j].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        lpPos[m].Followpos = NewHead;
                    }

                    j = lpPool[j].Next;
                }

                i = lpPool[i].Next;
            }
            break;

        case OR:
            if(LexTrace) if(TraceStr("| ") == FALSE) return FALSE;
            lpStack[StackIndex - 1].Nullable |= lpStack[StackIndex].Nullable;

            i = lpStack[StackIndex].Firstpos;
            while(i)
            {
                Same = FALSE;
                p = lpStack[StackIndex - 1].Firstpos;
                while(p)
                {
                    if(lpPool[p].Pos == lpPool[i].Pos) Same = TRUE;
                    p = lpPool[p].Next;
                }

                if(Same == FALSE)
                {
                    if(PoolAddPosToList(
                    lpStack[StackIndex - 1].Firstpos,
                    lpPool[i].Pos,
                    (LPDWORD)&NewHead)
                    == FALSE)
                    return FALSE;
                    lpStack[StackIndex - 1].Firstpos = NewHead;
                }

                i = lpPool[i].Next;
            }

            i = lpStack[StackIndex].Lastpos;
            while(i)
            {
                Same = FALSE;
                p = lpStack[StackIndex - 1].Lastpos;
                while(p)
                {
                    if(lpPool[p].Pos == lpPool[i].Pos) Same = TRUE;
                    p = lpPool[p].Next;
                }

                if(Same == FALSE)
                {
                    if(PoolAddPosToList(lpStack[StackIndex - 1].Lastpos, lpPool[i].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                    lpStack[StackIndex - 1].Lastpos = NewHead;
                }

                i = lpPool[i].Next;
            }

            POP;
            break;

        case CAT:
            if(LexTrace) if(TraceStr(". ") == FALSE) return FALSE;
            i = lpStack[StackIndex - 1].Lastpos;
            while(i)
            {
                m = lpPool[i].Pos;
                j = lpStack[StackIndex].Firstpos;
                while(j)
                {
                    Same = FALSE;
                    p = lpPos[m].Followpos;
                    while(p)
                    {
                        if(lpPool[p].Pos == lpPool[j].Pos) Same = TRUE;
                        p = lpPool[p].Next;
                    }

                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(lpPos[m].Followpos, lpPool[j].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        lpPos[m].Followpos = NewHead;
                    }

                    j = lpPool[j].Next;
                }

                i = lpPool[i].Next;
            }

            if(lpStack[StackIndex - 1].Nullable)
            {
                i = lpStack[StackIndex].Firstpos;
                while(i)
                {
                    Same = FALSE;
                    p = lpStack[StackIndex - 1].Firstpos;
                    while(p)
                    {
                        if(lpPool[p].Pos == lpPool[i].Pos) Same = TRUE;
                        p = lpPool[p].Next;
                    }

                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(lpStack[StackIndex - 1].Firstpos, lpPool[i].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        lpStack[StackIndex - 1].Firstpos = NewHead;
                    }

                    i = lpPool[i].Next;
                }
            }

            if(lpStack[StackIndex].Nullable)
            {
                i = lpStack[StackIndex].Lastpos;
                while(i)
                {
                    Same = FALSE;
                    p = lpStack[StackIndex - 1].Lastpos;
                    while(p)
                    {
                        if(lpPool[p].Pos == lpPool[i].Pos) Same = TRUE;
                        p = lpPool[p].Next;
                    }

                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(lpStack[StackIndex - 1].Lastpos, lpPool[i].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        lpStack[StackIndex - 1].Lastpos = NewHead;
                    }

                    i = lpPool[i].Next;
                }
            }
            else
            {
                lpStack[StackIndex - 1].Lastpos = lpStack[StackIndex].Lastpos;
                lpStack[StackIndex].Lastpos = 0;
            }

            lpStack[StackIndex - 1].Nullable &= lpStack[StackIndex].Nullable;

            POP;
            break;

        default:
            if(LexTrace)
            {
                if(LookAhead >= ' ' && LookAhead < MAX_CHARS)
                {
                    wsprintf((LPSTR)OutBuffer,"Leaf: %c ",(unsigned char)LookAhead);
                    if(TraceStr((LPSTR)OutBuffer) == FALSE) return FALSE;
                }
                else
                {
                    wsprintf((LPSTR)OutBuffer,"Leaf: @ (code=%u) ",LookAhead);
                    if(TraceStr((LPSTR)OutBuffer) == FALSE) return FALSE;
                }
            }

            /* Leaf node (for a character, a class code or an empty) */
            if(PushPosition() == FALSE) return FALSE;
            if(LookAhead == EMPTY)
            {
                lpPos[nPos].Label = EMPTY;
                lpStack[StackIndex].Nullable = TRUE;
                lpStack[StackIndex].Firstpos = 0;
                lpStack[StackIndex].Lastpos = 0;
            }
            else
            {
                /* Character or class code */
                lpPos[nPos].Label = c;
                for(i = 1; i <= Alphabet[0]; i++)
                    if(LookAhead == Alphabet[i]) goto NOT_ADD;
                Alphabet[0]++; Alphabet[Alphabet[0]] = LookAhead;

                NOT_ADD:
                lpStack[StackIndex].Nullable = FALSE;
                if(PoolAddPosToList(lpStack[StackIndex].Firstpos, nPos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                lpStack[StackIndex].Firstpos = NewHead;
                if(PoolAddPosToList(lpStack[StackIndex].Lastpos, nPos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                lpStack[StackIndex].Lastpos = NewHead;
            }

            if(LexTrace)
            {
                wsprintf((LPSTR)OutBuffer,",Position=%lu",nPos);
                if(TraceStr((LPSTR)OutBuffer) == FALSE) return FALSE;
            }
            break;
    }
    if(LexTrace) if(TracePos() == FALSE) return FALSE;

    return TRUE;
}

/* Read the next source token in the input buffer containing the current regular expression,
   and store it into LookAhead */
BOOL GetSourceToken()
{
    WORD c1;
    static WORD Length;
    static char Bc[4];      // For escape sequences

    /* Quoted strings: backslash used only to escape " and \ */
    if(lpRg[LookAheadIndex] == '"')
    {
        if(str == FALSE && lpRg[LookAheadIndex + 1] == '"') {LexErrorCode = LEXERR_EMPTY_STRING; return FALSE;}

        /* Enclose strings into parenthesis */
        if(str) {str = FALSE; LookAheadIndex++; LookAhead = RIGHT; return TRUE;}
        else {str = TRUE; LookAheadIndex++; LookAhead = LEFT; return TRUE;}
    }

    if(str)
    {
        c1 = lpRg[LookAheadIndex];
        if(c1 == '\\' && (lpRg[LookAheadIndex + 1] == '"' ||
                          lpRg[LookAheadIndex + 1] == '\\'))    // Beware of \" or \\ inside the string
        {
            LookAheadIndex++;
            LookAhead = lpRg[LookAheadIndex];
        }
        else LookAhead = c1;    // All other characters read literally
    }
    else
    {
        /* EMPTY character */
        if(suivant == EMPTY)
        {
            suivant = EOI;
            LookAhead = EMPTY;
            return TRUE;        // LookAheadIndex deliberately not incremented before return
        }
        else
        {
            switch(lpRg[LookAheadIndex])
            {
                case '\\':  // Backslash single character quotation not in string, not in class: may escape as in C
                    Bc[0] = '\\';
                    Bc[1] = (unsigned char)lpRg[LookAheadIndex + 1];
                    if(Bc[1])
                    {
                        Bc[2] = (unsigned char)lpRg[LookAheadIndex + 2];
                        if(Bc[2]) Bc[3] = (unsigned char)lpRg[LookAheadIndex + 3];
                    }
                    LookAhead = (WORD)BackslashConstant((LPHUGE)Bc,(LPWORD)&Length);
                    LookAheadIndex += Length - 1;
                    break;
                case '|':
                    LookAhead = OR;
                    c1 = lpRg[LookAheadIndex + 1];
                    switch(c1)
                    {
                        case '*':   LexErrorCode = LEXERR_OR_STAR; return FALSE;
                        case '?':   LexErrorCode = LEXERR_OR_QMARK; return FALSE;
                        case '+':   LexErrorCode = LEXERR_OR_PLUS; return FALSE;
                    }
                    if(c1 == '|' || c1 == ')' || c1 == EOI) suivant = EMPTY;    // Generate EMPTY character
                    break;
                case '*':
                    LookAhead = STAR;
                    switch(lpRg[LookAheadIndex + 1])
                    {
                        case '*':   LexErrorCode = LEXERR_STAR_STAR; return FALSE;
                        case '?':   LexErrorCode = LEXERR_STAR_QMARK; return FALSE;
                        case '+':   LexErrorCode = LEXERR_STAR_PLUS; return FALSE;
                    }
                    break;
                case '(':
                    LookAhead = LEFT;
                    c1 = lpRg[LookAheadIndex + 1];
                    if(c1 == ')') {LexErrorCode = LEXERR_MISSING_SUBEXPRESSION; return FALSE;}
                    if(c1 == '|') suivant = EMPTY;  // Generate EMPTY character
                    break;
                case ')':
                    LookAhead = RIGHT;
                    break;
                case '+':
                    LookAhead = PLUS;
                    switch(lpRg[LookAheadIndex + 1])
                    {
                        case '*':   LexErrorCode = LEXERR_PLUS_STAR; return FALSE;
                        case '?':   LexErrorCode = LEXERR_PLUS_QMARK; return FALSE;
                        case '+':   LexErrorCode = LEXERR_PLUS_PLUS; return FALSE;
                    }
                    break;
                case '?':
                    LookAhead = QMARK;
                    switch(lpRg[LookAheadIndex + 1])
                    {
                        case '*':   LexErrorCode = LEXERR_QMARK_STAR; return FALSE;
                        case '?':   LexErrorCode = LEXERR_QMARK_QMARK; return FALSE;
                        case '+':   LexErrorCode = LEXERR_QMARK_PLUS; return FALSE;
                    }
                    break;
                default:    // A character or a class code!
                    LookAhead = lpRg[LookAheadIndex];
                    break;
            }
        }
    }

    LookAheadIndex++;

    return TRUE;
}

/* Build state and transition tables for DFA of current regular expression */
BOOL Tables()
{
    register DWORD fpos,i,j,k,p,r,m,q;
    register BOOL Same;
    static DWORD NewHead;

    if(AddState() == FALSE) return FALSE;
    if(AddTransitionG() == FALSE) return FALSE;
    lpStates[1].Subset = lpStack[StackIndex].Firstpos;
    lpStack[StackIndex].Firstpos = 0;
    hpGNextState[TotalStates + 1].Group = 1;    // Mark group position on first state

    for(i = 1; i <= nStates; i++)               // Loop on states already in the state table
    {
        for(j = 1; j <= Alphabet[0]; j++)       // Loop on characters in the alphabet and class codes
        {
            u = 0;                              // Empty subset
            k = lpStates[i].Subset;
            while(k)                            // Loop on positions in current state
            {
                q = lpPool[k].Pos;
                if(lpPos[q].Label == Alphabet[j])
                {
                    p = lpPos[q].Followpos;
                    while(p)                    // Loop on the positions in Followpos(current position)
                    {
                        fpos = lpPool[p].Pos;
                        Same = FALSE;
                        r = u;
                        while(r)                // Loop on positions already in the built subset
                        {
                            if((WORD)lpPool[r].Pos == fpos) Same = TRUE;    // Position already in subset u
                            r = lpPool[r].Next;
                        }
    
                        if(Same == FALSE)
                        {
                            if(PoolAddPosToList(u, lpPool[p].Pos, (LPDWORD)&NewHead) == FALSE) return FALSE;
                            u = NewHead;
                        }

                        p = lpPool[p].Next;
                    }
                }

                k = lpPool[k].Next;
            }

            /* The subset u is built */
            Same = FALSE;
            if(u == 0) continue;            // Empty subset: try next character!
            for(k = 1; k <= nStates; k++)
            {
                if(PoolListLength(lpStates[k].Subset) != PoolListLength(u)) continue;
                p = lpStates[k].Subset;
                while(p)
                {
                    fpos = lpPool[p].Pos;
                    Same = FALSE;
                    r = u;
                    while(r)
                    {
                        if(fpos == (WORD)lpPool[r].Pos) Same = TRUE;
                        r = lpPool[r].Next;
                    }
                    if(Same == FALSE) break;
    
                    p = lpPool[p].Next;
                }
                if(Same == TRUE) break;
            }
            if(Same) {m = k; PoolFreeList(u); u = 0;}
            else
            {
                /* Add new state to state table; add entry to next state table for this state */
                if(AddState() == FALSE) return FALSE;
                lpStates[nStates].Subset = u;
                u = 0;
                m = nStates;
            }
            if(AddTransitionG() == FALSE) return FALSE;
            if(Alphabet[j] < CLASS_CODES)
                hpGNextState[TotalStates+(WORD)i].NextState[Alphabet[j]]
                    = (WORD)m;          // Transition on a character
            else
                hpGNextState[TotalStates+(WORD)i].NextStateClass[Alphabet[j] - CLASS_CODES]
                    = (WORD)m;          // Transition on a character class!
        }
    }

    return TRUE;
}

/* Add token numbers to appropriate states */
BOOL AddToken()
{
    register DWORD i,j;

    /* Create a token number for states whose set of positions contains the endmarker (last position).
       The token number is saved in the transition table. */
    nToken++;
    for(i = 1; i <= nStates; i++)
    {
        j = lpStates[i].Subset;
        while(j)
        {
            if((DWORD)lpPool[j].Pos == nPos) {hpGNextState[TotalStates+i].Token = nToken; break;}
            j = lpPool[j].Next;
        }
    }

    return TRUE;
}

/* Clear variables before processing next regular expression */
BOOL Clear()
{
    register WORD i;

    nPool = 0;
    Holes = 0;
    TotalStates += nStates;
    for(i = 1; i <= Alphabet[0]; i++) Alphabet[i] = 0;
    Alphabet[0] = 0;
    nPos = 0;
    u = 0;
    StackIndex = 0;
    str = FALSE;
    suivant = 0;
    nStates = 0;

    return TRUE;
}

/* Construct global NFA, by creating a new start state composed of start states of
   each DFA. Then apply a subset construction to get final global DFA.
   The new state table uses the lpStates[] table again.
   Note that the global NFA has no EMPTY transition, since it is composed of DFAs.
   Consequently, there is no need for an empty-closure function. */
BOOL AttachDFAs()
{          
    register WORD i,j,l,m;
    register DWORD k,p,r;
    WORD Offset;
    BOOL Same;
    static DWORD NewHead;


    /* Add the appropriate offset to each transition value */
    for(i = 1; i <= TotalStates; i++)
    {
        if(hpGNextState[i].Group == 1) Offset = (WORD)(i - 1);
        for(j = 1; j < MAX_CHARS; j++)
            if(hpGNextState[i].NextState[j]) hpGNextState[i].NextState[j] += Offset;
        for(j = 0; j < nClass; j++)
            if(hpGNextState[i].NextStateClass[j]) hpGNextState[i].NextStateClass[j] += Offset;
    }

    /* Build start state of global NFA */
    if(AddState() == FALSE) return FALSE;
    for(i = 1; i <= TotalStates; i++)
    {
        if(hpGNextState[i].Group == 1)  // Found the start state of a DFA
        {
            hpGNextState[i].Group = 0;

            /* Add the start state of the DFA to the new start state */
            if(PoolAddPosToList(lpStates[1].Subset, i, (LPDWORD)&NewHead) == FALSE) return FALSE;
            lpStates[1].Subset = NewHead;

            /* Token: the smallest token value among the start states of DFAs, represents the regular
               expression listed first in the lex source input. */
            l = lpStates[1].Token; m = hpGNextState[i].Token;
            if(l == 0) lpStates[1].Token = m;
            else lpStates[1].Token = ((m && m < l) ? m : l);
        }
    }

    for(i = 1; i <= nStates; i++)           // Loop on states already in the state table
    {
        for(j = 0; j < MAX_CHARS; j++)      // Loop on all characters
        {
            u = 0;                          // Empty subset
            k = lpStates[i].Subset;
            while(k)                        // Loop on substates in current state
            {
                m = hpGNextState[lpPool[k].Pos].NextState[j];
                if(m)                       // Transition on current character ?
                {
                    Same = FALSE;
                    p = u;
                    while(p)
                    {
                        if((WORD)lpPool[p].Pos == m) {Same = TRUE; break;}  // Substate already in subset u
                        p = lpPool[p].Next;
                    }
    
                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(u, m, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        u = NewHead;
                    }
                }

                /* Consider possible transitions due to classes */
                for(r = 0; r < nClass; r++) // Loop on classes
                {
                    m = hpGNextState[lpPool[k].Pos].NextStateClass[r];
                    if(m && hpClass[r][j] == 1) // There is a transition on the class, and the class contains the character j!
                    {
                        Same = FALSE;
                        p = u;
                        while(p)
                        {
                            if((WORD)lpPool[p].Pos == m) {Same = TRUE; break;}  // Substate already in subset u
                            p = lpPool[p].Next;
                        }
        
                        if(Same == FALSE)
                        {
                            if(PoolAddPosToList(u, m, (LPDWORD)&NewHead) == FALSE) return FALSE;
                            u = NewHead;
                        }
                    }
                }

                k = lpPool[k].Next;
            }

            /* The subset u is built: is state u already in lpStates ? */
            if(u == 0) continue;        // Empty subset: try next character!
            Same = FALSE;
            for(k = 1; k <= nStates; k++)
            {
                if(PoolListLength(lpStates[k].Subset) != PoolListLength(u)) continue;
                p = lpStates[k].Subset;
                while(p)
                {
                    Same = FALSE;
                    r = u;
                    while(r)
                    {
                        if(lpPool[p].Pos == lpPool[r].Pos) {Same = TRUE; break;}
                        r = lpPool[r].Next;
                    }
                    if(Same == FALSE) break;
                    p = lpPool[p].Next;
                }
                if(Same == TRUE) break;
            }
            if(Same) {m = (WORD)k; PoolFreeList(u); u = 0;}
            else
            {
                /* Add new state to state table; add entry to final next state table for this state */
                if(AddState() == FALSE) return FALSE;
                lpStates[nStates].Subset = u;
                k = u;
                u = 0;
                while(k)
                {
                    /* Token: the smallest token value among the start states of DFAs, represents the regular
                       expression listed first in the lex source input. */
                    l = lpStates[nStates].Token; m = hpGNextState[lpPool[k].Pos].Token;
                    if(l == 0) lpStates[nStates].Token = m;
                    else lpStates[nStates].Token = ((m && m < l) ? m : l);
    
                    m = nStates;
                    k = lpPool[k].Next;
                }
            }
            if(AddTransitionF() == FALSE) return FALSE;
            hpFNextState[i].NextState[j] = m;
        }

        /* Add transitions for classes */
        for(j = 0; j < nClass; j++)         // Loop on all classes
        {
            u = 0;                          // Empty subset
            k = lpStates[i].Subset;
            while(k)                        // Loop on substates in current state
            {
                m = hpGNextState[lpPool[k].Pos].NextStateClass[j];
                if(m)                       // Transition on current class ?
                {
                    Same = FALSE;
                    p = u;
                    while(p)
                    {
                        if((WORD)lpPool[p].Pos == m) {Same = TRUE; break;}  // Substate already in subset u
                        p = lpPool[p].Next;
                    }
    
                    if(Same == FALSE)
                    {
                        if(PoolAddPosToList(u, m, (LPDWORD)&NewHead) == FALSE) return FALSE;
                        u = NewHead;
                    }
                }

                k = lpPool[k].Next;
            }

            /* The subset u is built: is state u already in lpStates ? */
            if(u == 0) continue;        // Empty subset: try next character!
            Same = FALSE;
            for(k = 1; k <= nStates; k++)
            {
                if(PoolListLength(lpStates[k].Subset) != PoolListLength(u)) continue;
                p = lpStates[k].Subset;
                while(p)
                {
                    Same = FALSE;
                    r = u;
                    while(r)
                    {
                        if(lpPool[p].Pos == lpPool[r].Pos) {Same = TRUE; break;}
                        r = lpPool[r].Next;
                    }
                    if(Same == FALSE) break;
                    p = lpPool[p].Next;
                }
                if(Same == TRUE) break;
            }
            if(Same) {m = (WORD)k; PoolFreeList(u); u = 0;}
            else
            {
                /* Add new state to state table; add entry to final next state table for this state */
                if(AddState() == FALSE) return FALSE;
                lpStates[nStates].Subset = u;
                k = u;
                u = 0;
                while(k)
                {
                    /* Token: the smallest token value among the start states of DFAs, represents the regular
                       expression listed first in the lex source input. */
                    l = lpStates[nStates].Token; m = hpGNextState[lpPool[k].Pos].Token;
                    if(l == 0) lpStates[nStates].Token = m;
                    else lpStates[nStates].Token = ((m && m < l) ? m : l);
    
                    m = nStates;
                    k = lpPool[k].Next;
                }
            }
            if(AddTransitionF() == FALSE) return FALSE;
            hpFNextState[i].NextStateClass[j] = m;
        }
    }

    return TRUE;
}

/* Transform transitions on classes into transitions on characters */
void AddClassTransitions()
{
    register WORD i,j,k,m;

    for(i = 1; i <= TotalStates; i++)       // Loop on new states
    {
        for(j = 0; j < nClass; j++)         // Loop on classes
        {
            for(k = 1; k < MAX_CHARS; k++)  // Loop on characters
            {
                m = hpGNextState[i].NextStateClass[j];
                if(m &&                                             // There is a transition on state i, class j
                   hpClass[j][k] == 1 &&                            // Class j contains character k
                   hpGNextState[i].NextState[k] == 0)               // There is no transition on state i, character k
                        hpGNextState[i].NextState[k] = m;
            }
        }
    }
}