/******************************************************************
* Phoenix Lexical Analyzer Generator.                             *
* Management of dynamic arrays.                                   *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"
#include "ph_parse.h"
#include "ph_error.h"

/* Increment the top of the pool, and store a position on the new top. */
BOOL PoolPush(DWORD Position)
{
    BOOL NewMemory;

    NewMemory = FALSE;

    if(lpPool == NULL)
    {
        // Allocate Space for array
        nMaxPool = POOL_INCREMENT;
        lpPool = (LPPOOL)GlobalAllocPtr(GHND,(DWORD)sizeof(POOL) * nMaxPool);
        nPool = 0;
        NewMemory = TRUE;
    }
    else
        if(nPool + 1 >= nMaxPool - 1)
        {
            // Allocate extra Space for array
            nMaxPool += POOL_INCREMENT;
            lpPool = (LPPOOL)GlobalReAllocPtr(lpPool,(DWORD)sizeof(POOL) * nMaxPool,GMEM_ZEROINIT|GMEM_MOVEABLE);
            NewMemory = TRUE;
        }

    if(NewMemory && lpPool == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    lpPool[++nPool].Pos = Position;

    return TRUE;
}

/* Increment the top of the stack. */
BOOL PushStack()
{
    BOOL NewMemory;

    NewMemory = FALSE;

    if(lpStack == NULL)
    {
        // Allocate Space for array
        nMaxStack = STACK_INCREMENT;
        lpStack = (LPSTACK)GlobalAllocPtr(GHND,(DWORD)sizeof(STACK) * (DWORD)nMaxStack);
        StackIndex = 0;
        NewMemory = TRUE;
    }
    else
        if(StackIndex + 1 >= nMaxStack - 1)
        {
            // Allocate extra Space for array
            nMaxStack += STACK_INCREMENT;
            lpStack = (LPSTACK)GlobalReAllocPtr(lpStack,(DWORD)sizeof(STACK) * (DWORD)nMaxStack,GMEM_ZEROINIT|GMEM_MOVEABLE);
            NewMemory = TRUE;
        }

    if(NewMemory && lpStack == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    StackIndex++;
    lpStack[StackIndex].Nullable = FALSE;
    lpStack[StackIndex].Firstpos = 0;
    lpStack[StackIndex].Lastpos = 0;

    return TRUE;
}

/* Increment the top of the position array. */
BOOL PushPosition()
{
    BOOL NewMemory;

    NewMemory = FALSE;

    if(lpPos == NULL)
    {
        // Allocate Space for array
        nMaxPos = POSITION_INCREMENT;
        lpPos = (LPPOSITION)GlobalAllocPtr(GHND,(DWORD)sizeof(POSITION) * (DWORD)nMaxPos);
        nPos = 0;
        NewMemory = TRUE;
    }
    else
        if(nPos + 1 >= nMaxPos - 1)
        {
            // Allocate extra Space for array
            nMaxPos += POSITION_INCREMENT;
            lpPos = (LPPOSITION)GlobalReAllocPtr(lpPos,(DWORD)sizeof(POSITION) * (DWORD)nMaxPos,GMEM_ZEROINIT|GMEM_MOVEABLE);
            NewMemory = TRUE;
        }

    if(NewMemory && lpPos == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    nPos++;
    lpPos[nPos].Label = '\0';
    lpPos[nPos].Followpos = 0;

    return TRUE;
}

/* Allocate hpClass */
BOOL AllocateClass()
{
  if(hpClass == NULL)
  {
    // Allocate Space for array
    hpClass = (LPCLASS)GlobalAllocPtr(GHND,(DWORD)sizeof(ONE_CLASS) * (DWORD)MAX_CLASSES);
    nClass = 0;

    if(hpClass == NULL)
    {
      LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
      return FALSE;
    }
  }

  return TRUE;
}

/* Add new state to state table */
BOOL AddState()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(lpStates == NULL)
  {
    // Allocate Space for array
    nMaxStates = LEXSTATE_INCREMENT;
    lpStates = (LPLEXSTATE)GlobalAllocPtr(GHND,(DWORD)sizeof(LEXSTATE) * (DWORD)nMaxStates);
    nStates = 0;
    NewMemory = TRUE;
  }
  else
    if(nStates + 1 >= nMaxStates - 1)   // Since index 0 is not used!
    {
        // Allocate extra Space for array
        nMaxStates += LEXSTATE_INCREMENT;
        lpStates = (LPLEXSTATE)GlobalReAllocPtr(lpStates,(DWORD)sizeof(LEXSTATE) * (DWORD)nMaxStates,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && lpStates == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }

  nStates++;    // The index of the last state, since index 0 is not used!
  lpStates[nStates].Subset = 0;
  lpStates[nStates].Token = 0;

  return TRUE;
}

/* Add new transition to hpGNextState table */
BOOL AddTransitionG()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpGNextState == NULL)
  {
    // Allocate Space for array
    nMaxGNextState = LEXTRANSITION_INCREMENT;
    hpGNextState = (LPLEXTRANSITION)GlobalAllocPtr(GHND,(DWORD)sizeof(LEXTRANSITION) * (DWORD)nMaxGNextState);
    NewMemory  = TRUE;
  }
  else
    if(nStates + TotalStates + 1 >= nMaxGNextState - 1) // Since index 0 is not used!
    {
        // Allocate extra Space for array
        nMaxGNextState += LEXTRANSITION_INCREMENT;
        hpGNextState = (LPLEXTRANSITION)GlobalReAllocPtr(hpGNextState,(DWORD)sizeof(LEXTRANSITION) * (DWORD)nMaxGNextState,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory)
  {
    if(hpGNextState == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }
  }

  return TRUE;
}

/* Add new transition to hpFNextState final transition table */
BOOL AddTransitionF()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpFNextState == NULL)
  {
    // Allocate Space for array
    nMaxFNextState = LEXTRANSITION_INCREMENT;
    hpFNextState = (LPLEXTRANSITION)GlobalAllocPtr(GHND,(DWORD)sizeof(LEXTRANSITION) * (DWORD)nMaxFNextState);
    NewMemory  = TRUE;
  }
  else
    if(nStates + 1 >= nMaxFNextState - 1)   // Since index 0 is not used!
    {
        // Allocate extra Space for array
        nMaxFNextState += LEXTRANSITION_INCREMENT;
        hpFNextState = (LPLEXTRANSITION)GlobalReAllocPtr(hpFNextState,(DWORD)sizeof(LEXTRANSITION) * (DWORD)nMaxFNextState,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && hpFNextState == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }

  return TRUE;
}

/* Adds a regular expression to the reg. exp. list of Part 2 */
BOOL AddP2RegularExpression()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpP2RegExp == NULL)
  {
    // Allocate Space for array of regular expressions
    nMaxP2RegExp = LEXP2REGEXP_INCREMENT;
    hpP2RegExp = (HPLEXP2REGEXP)GlobalAllocPtr(GHND,(DWORD)sizeof(LEXP2REGEXP) * (DWORD)nMaxP2RegExp);
    nP2RegExp = 0;
    NewMemory = TRUE;
  }
  else
    if(nP2RegExp + 1 >= nMaxP2RegExp - 1)
    {
        // Allocate extra Space for array of regular expressions
        nMaxP2RegExp += LEXP2REGEXP_INCREMENT;
        hpP2RegExp = (HPLEXP2REGEXP)GlobalReAllocPtr(hpP2RegExp,(DWORD)sizeof(LEXP2REGEXP) * (DWORD)nMaxP2RegExp,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && hpP2RegExp == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }
  nP2RegExp++;

  return TRUE;
}

/* Adds a declaration to the decl. list of Part 1 (identifier regexp). */
BOOL AddP1Declaration()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(lpP1Decl == NULL)
  {
    // Allocate Space for array of declarations
    nMaxP1Decl = LEXP1DECL_INCREMENT;
    lpP1Decl = (LPLEXP1DECL)GlobalAllocPtr(GHND,(DWORD)sizeof(LEXP1DECL) * (DWORD)nMaxP1Decl);
    nP1Decl = 0;
    NewMemory = TRUE;
  }
  else
    if(nP1Decl + 1 >= nMaxP1Decl - 1)
    {
        // Allocate extra Space for array of declarations
        nMaxP1Decl += LEXP1DECL_INCREMENT;
        lpP1Decl = (LPLEXP1DECL)GlobalReAllocPtr(lpP1Decl,(DWORD)sizeof(LEXP1DECL) * (DWORD)nMaxP1Decl,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && lpP1Decl == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }
  nP1Decl++;

  return TRUE;
}

/* Add a symbol to the symbol table */
BOOL AddSymbol()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpSymbol == NULL)
  {
    // Allocate Space for array of symbols
    nMaxSymbol = SYMBOL_INCREMENT;
    hpSymbol = (LPSYMBOL)GlobalAllocPtr(GHND,(DWORD)sizeof(SYMBOLS) * (DWORD)nMaxSymbol);
    nSymbol = 0;
    NewMemory = TRUE;
  }
  else
    if(nSymbol + 1 >= nMaxSymbol - 1)
    {
        // Allocate extra Space for array of symbols
        nMaxSymbol += SYMBOL_INCREMENT;
        hpSymbol = (LPSYMBOL)GlobalReAllocPtr(hpSymbol,(DWORD)sizeof(SYMBOLS) * (DWORD)nMaxSymbol,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && hpSymbol == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }

  nSymbol++;

  return TRUE;
}

/* Add a production to the production table */
BOOL AddProduction()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpProduction == NULL)
  {
    // Allocate Space for array of productions
    nMaxProduction = PRODUCTION_INCREMENT;
    hpProduction = (LPPRODUCTION)GlobalAllocPtr(GHND,(DWORD)sizeof(PRODUCTION) * (DWORD)nMaxProduction);
    nProduction = 0;
    NewMemory = TRUE;
  }
  else
    if(nProduction + 1 >= nMaxProduction - 1)
    {
        // Allocate extra Space for array of productions
        nMaxProduction += PRODUCTION_INCREMENT;
        hpProduction = (LPPRODUCTION)GlobalReAllocPtr(hpProduction,(DWORD)sizeof(PRODUCTION) * (DWORD)nMaxProduction,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && hpProduction == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }

  nProduction++;

  return TRUE;
}

/* Add a set of items to the collection of sets of items table hpCsiTable */
BOOL AddCsi()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpCsiTable == NULL)
  {
    // Allocate Space for array of sets of items
    nMaxCsi = CSI_INCREMENT;
    hpCsiTable = (LPSITEMS)GlobalAllocPtr(GHND,(DWORD)sizeof(SET_OF_ITEMS) * (DWORD)nMaxCsi);
    nCsi = 0;
    NewMemory = TRUE;
  }
  else
    if(nCsi + 1 >= nMaxCsi - 1)
    {
        // Allocate extra Space for array of sets of items
        nMaxCsi += CSI_INCREMENT;
        hpCsiTable = (LPSITEMS)GlobalReAllocPtr(hpCsiTable,(DWORD)sizeof(SET_OF_ITEMS) * (DWORD)nMaxCsi,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && hpCsiTable == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }

  nCsi++;

  return TRUE;
}

/* Add an entry to the action-goto table */
BOOL AddActionGoto()
{
  register WORD i,j;

  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpActionGoto == NULL)
  {
    // Allocate Space for array of action-gotos
    nMaxActionGoto = ACTION_GOTO_INCREMENT;
    hpActionGoto = (LPACTIONGOTO)GlobalAllocPtr(GHND,(DWORD)sizeof(short int) * (DWORD)nMaxActionGoto * (DWORD)nSymbol);
    nActionGoto = 0;
    NewMemory = TRUE;
  }
  else
    if(nActionGoto + 1 >= nMaxActionGoto - 1)
    {
        // Allocate extra Space for array of action-gotos
        nMaxActionGoto += ACTION_GOTO_INCREMENT;
        hpActionGoto = (LPACTIONGOTO)GlobalReAllocPtr(hpActionGoto,(DWORD)sizeof(short int) * (DWORD)nMaxActionGoto * (DWORD)nSymbol,
            GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory)
      if(hpActionGoto == NULL)
      {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
      }
      else
      {
        /* Set new entries to blank */
        for(i = nMaxActionGoto - ACTION_GOTO_INCREMENT; i < nMaxActionGoto; i++)
            for(j = 0; j < nSymbol; j++) hpActionGoto[(DWORD)nSymbol * (DWORD)i + (DWORD)j] = BLANK;
      }

  nActionGoto++;

  return TRUE;
}

/* Add an entry to the propagation table */
BOOL AddPropa()
{
  BOOL NewMemory;

  NewMemory = FALSE;

  if(hpPropa == NULL)
  {
    // Allocate Space for array of productions
    nMaxPropa = PROPA_INCREMENT;
    hpPropa = (LPPROPA)GlobalAllocPtr(GHND,(DWORD)sizeof(PROPAGATION) * (DWORD)nMaxPropa);
    nPropa = 0;
    NewMemory = TRUE;
  }
  else
    if(nPropa + 1 >= nMaxPropa - 1)
    {
        // Allocate extra Space for array
        nMaxPropa += PROPA_INCREMENT;
        hpPropa = (LPPROPA)GlobalReAllocPtr(hpPropa,(DWORD)sizeof(PROPAGATION) * (DWORD)nMaxPropa,GMEM_ZEROINIT|GMEM_MOVEABLE);
        NewMemory = TRUE;
    }

  if(NewMemory && hpPropa == NULL)
  {
    LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
    return FALSE;
  }

  nPropa++;

  return TRUE;
}
