/******************************************************************
* Phoenix Lexical Analyzer Generator.                             *
* Packing the global DFA.                                         *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"
#include "ph_error.h"

static WORD Set[MAX_CHARS][2],nSet;

/* Insert a set of values (char,nextstate) in the next state array */
BOOL Insert(WORD State)
{
    register WORD i,j;
    BOOL Checked;

    j = MAX_CHARS;      // Skip the zone for state 0!
    while(1)
    {
        if(j + MAX_CHARS > nMaxCheck)
        {
            nMaxCheck += CHECK_INCREMENT;
            PCheck = (LPWORD)GlobalReAllocPtr(PCheck,(DWORD)sizeof(WORD) * (DWORD)nMaxCheck,GMEM_MOVEABLE);
            if(PCheck == NULL)
            {
                LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
                return FALSE;
            }
            for(i = nCheck; i < nMaxCheck; i++) PCheck[i] = VIRGIN;
            PNext = (LPWORD)GlobalReAllocPtr(PNext,(DWORD)sizeof(WORD) * (DWORD)nMaxCheck,GMEM_ZEROINIT|GMEM_MOVEABLE);
            if(PNext == NULL)
            {
                LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
                return FALSE;
            }
        }

        Checked = FALSE;
        for(i = 0; i < nSet; i++)
            if(PCheck[j + Set[i][1]] != VIRGIN) {j++; Checked = TRUE; break;}   // There is a difference between 0 and VIRGIN!

        if(Checked == FALSE) break;
    }

    PBase[State] = j;
    for(i = 0; i < nSet; i++)   // Loop on each character of the set
    {
        PNext[j + Set[i][1]] = Set[i][0];
        PCheck[j + Set[i][1]] = State;
    }
    if(j + MAX_CHARS > nCheck) nCheck = j + MAX_CHARS;

    return TRUE;
}

/* Build packed arrays for DFA */
BOOL PackDFA()
{
    register WORD i,j,k;
    WORD Like;
    BOOL Merged;

    nCheck = 0;
    nMaxCheck = CHECK_INCREMENT;
    PBase = (LPWORD)GlobalAllocPtr(GHND,(DWORD)sizeof(WORD) * (DWORD)(TotalStates + 1));
    if(PBase == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    PDefault = (LPWORD)GlobalAllocPtr(GHND,(DWORD)sizeof(WORD) * (DWORD)(TotalStates + 1));
    if(PDefault == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    PNext = (LPWORD)GlobalAllocPtr(GHND,(DWORD)sizeof(WORD) * (DWORD)nMaxCheck);
    if(PNext == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    PCheck = (LPWORD)GlobalAllocPtr(GMEM_MOVEABLE,(DWORD)sizeof(WORD) * (DWORD)nMaxCheck);
    if(PCheck == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }
    for(i = 0; i < nMaxCheck; i++) PCheck[i] = VIRGIN;

    /* State 0 */
    PDefault[0] = 0;
    PBase[0] = 0;
    for(k = 0; k < MAX_CHARS; k++)
    {
        PNext[k] = 0;
        PCheck[k] = 0;
    }

    for(i = 1; i <= TotalStates; i++)
    {
        Merged = FALSE;
        for(j = 1; j < i; j++)
        {
            Like = 0;
            for(k = 0; k < MAX_CHARS; k++)
                if(hpGNextState[i].NextState[k] == hpGNextState[j].NextState[k]) Like++;
            if(Like >= MAX_CHARS * Percent / 100)
            {
                /*** Merge! ***/
                PDefault[i] = j;
                nSet = 0;
                for(k = 0; k < MAX_CHARS; k++)
                {
                    if(hpGNextState[i].NextState[k] != hpGNextState[j].NextState[k])
                    {
                        Set[nSet][0] = hpGNextState[i].NextState[k];
                        Set[nSet][1] = k;
                        nSet++;
                    }
                }
                if(Insert(i) == FALSE) return FALSE;
                Merged = TRUE;
                break;
            }
        }

        if(Merged == FALSE)
        {
            for(k = 0; k < MAX_CHARS; k++)
            {
                Set[k][0] = hpGNextState[i].NextState[k];
                Set[k][1] = k;
            }
            nSet = MAX_CHARS;
            if(Insert(i) == FALSE) return FALSE;
        }
    }

    return TRUE;
}
