/******************************************************************
* Phoenix Parser Generator.                                       *
* Packing the action and goto tables.                             *
******************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"
#include "ph_parse.h"
#include "ph_error.h"

/* Insert an action state in AGPool */
BOOL InsertAction(WORD State,short int DefReduce)
{
    register WORD i;
    register DWORD lx;
    short int Action;

    /* Make sure there is room for the complete state */
    while(nAGPool + nTerm > nMaxAGPool - 1)
    {
        nMaxAGPool += AGPOOL_INCREMENT;
        AGPool = (DWORD HOST_HUGE *)GlobalReAllocPtr(AGPool,(DWORD)sizeof(DWORD) * (DWORD)nMaxAGPool,GMEM_MOVEABLE);
        if(AGPool == NULL)
        {
            LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
            return FALSE;
        }
    }

    lx = (DWORD)nSymbol * (DWORD)State;

    ActionState[State] = nAGPool;

    for(i = 0; i < nTerm; i++)  // Loop on symbols
    {
        if((Action = hpActionGoto[lx + (DWORD)i]) != DefReduce && Action != BLANK)
            AGPool[nAGPool++] = MAKELONG(i,(UINT)Action);
    }
    AGPool[nAGPool++] = (DWORD)MAKELONG((UINT)END_OF_LIST,(UINT)DefReduce); // End of list

    return TRUE;
}

/* Insert a goto column in AGPool */
BOOL InsertGoto(WORD Nonterminal,WORD DefaultState)
{
    register WORD i;
    short int Go_to;

    /* Make sure there is room for the complete column */
    while(nAGPool + nCsi > nMaxAGPool - 1)
    {
        nMaxAGPool += AGPOOL_INCREMENT;
        AGPool = (DWORD HOST_HUGE *)GlobalReAllocPtr(AGPool,(DWORD)sizeof(DWORD) * (DWORD)nMaxAGPool,GMEM_MOVEABLE);
        if(AGPool == NULL)
        {
            LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
            return FALSE;
        }
    }

    GotoSymbol[Nonterminal - nTerm] = nAGPool;      // Note that GotoSymbol[0] is unused!

    for(i = 0; i < nCsi; i++)   // Loop on states
    {
        if((Go_to = hpActionGoto[(DWORD)nSymbol * (DWORD)i + (DWORD)Nonterminal]) != BLANK && (WORD)Go_to != DefaultState)
            AGPool[nAGPool++] = MAKELONG(i,(UINT)Go_to);
    }
    AGPool[nAGPool++] = MAKELONG((UINT)END_OF_LIST,(UINT)DefaultState); // End of list

    return TRUE;
}

/* Build packed arrays for Action and Goto tables */
BOOL PackActionGoto()
{
    register WORD i,j,k,m;
    register DWORD lx,lx2;
    short int Go_to,DefaultReduction,Value;
    WORD nShift,n;
    BOOL CompressState;

    nAGPool = 0;
    nMaxAGPool = AGPOOL_INCREMENT;
    AGPool = (DWORD HOST_HUGE *)GlobalAllocPtr(GHND,(DWORD)sizeof(DWORD) * (DWORD)nMaxAGPool);
    if(AGPool == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    ActionState = (short int HOST_FAR *)GlobalAllocPtr(GHND,(DWORD)sizeof(short int) * (DWORD)nCsi);
    if(ActionState == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    GotoSymbol = (LPWORD)GlobalAllocPtr(GHND,(DWORD)sizeof(WORD) * (DWORD)(nSymbol - nTerm));
    if(GotoSymbol == NULL)
    {
        LexErrorCode = LEXERR_MEMORY_ALLOC_ERROR;
        return FALSE;
    }

    /*** ActionState ***/

    /* State 0 */
    if(InsertAction(0,(short int)BLANK) == FALSE) return FALSE;

    /* Other states */
    for(i = 1; i < nCsi; i++)               // Loop on states
    {
        lx = (DWORD)nSymbol * (DWORD)i;

        for(j = 0; j < i; j++)              // Loop on previously processed states, look for matching state
        {
            lx2 = (DWORD)nSymbol * (DWORD)j;

            for(k = 0; k < nTerm; k++)  // Loop on symbols
                if(hpActionGoto[lx + (DWORD)k] != hpActionGoto[lx2 + (DWORD)k]) break;  // No match!

            if(k >= nTerm) break;       // Match!!!
        }

        if(j < i) ActionState[i] = ActionState[j];  // Merge state i into previous state j
        else
        {
            /* Try single reduction (all nonerror entries on the state are the same reduce action),
               or almost-single reduction (same as previously, except for 1 action) */
            DefaultReduction = BLANK;
            n = 0;      // Number of non-blank entries
            CompressState = TRUE;
            nShift = 0; // Number of shift actions
            for(j = 0; j < nTerm; j++)
            {
                Value = hpActionGoto[lx + (DWORD)j];
                if(Value != BLANK)
                {
                    n++;
                    if(Value < 0)   // Reduction
                        if(DefaultReduction == BLANK) DefaultReduction = Value;
                        else
                        {
                            if(Value != DefaultReduction) {CompressState = FALSE; break;}
                        }
                    else            // Shift
                        nShift++;
                }
            }

            if(DefaultReduction != BLANK && CompressState == TRUE && nShift == 0)   // Single reduction
                ActionState[i] = DefaultReduction;
            else
            {
                if(CompressState && nShift > n / 2) CompressState = FALSE;      // To compress, at least 50% of the same reduction!
    
                if(CompressState == FALSE) DefaultReduction = BLANK;
    
                if(InsertAction(i,DefaultReduction) == FALSE) return FALSE;
            }
        }
    }

    /*** GotoSymbol ***/

    for(i = nTerm + 1; i < nSymbol; i++)    // Loop on nonterminals
    {
        /* Count the occurrences of each nonterminal. We use the hpActionGoto[][S'] column, since it's unused! */
        for(j = 0; j < nCsi; j++) hpActionGoto[(DWORD)nSymbol * (DWORD)j + (DWORD)nTerm] = 0;   // Clear column
        for(j = 0; j < nCsi; j++)
            if((Go_to = hpActionGoto[(DWORD)nSymbol * (DWORD)j + (DWORD)i]) != BLANK)
                hpActionGoto[(DWORD)nSymbol * (DWORD)Go_to + (DWORD)nTerm]++;
        k = 0; m = 0;
        for(j = 0; j < nCsi; j++)
            if((WORD)(Go_to = hpActionGoto[(DWORD)nSymbol * (DWORD)j + (DWORD)nTerm]) > k) {k = Go_to; m = j;}

        if(InsertGoto(i,m) == FALSE) return FALSE;
    }

    return TRUE;
}

void FreePackedArrays()
{
  if(ActionState != NULL)
  {
    GlobalFreePtr(ActionState);
    ActionState = NULL;
  }

  if(GotoSymbol != NULL)
  {
    GlobalFreePtr(GotoSymbol);
    GotoSymbol = NULL;
  }

  if(AGPool != NULL)
  {
    GlobalFreePtr(AGPool);
    AGPool = NULL;
  }
}