/********************************************************************
* Phoenix Parser Generator.                                         *
* Reading and analyzing parser source file.                         *
********************************************************************/

#include <windows.h>
#include <windowsx.h>
#include "ph_scan.h"    // For input buffer declarations
#include "ph_parse.h"
#include "ph_error.h"

#define  up(c) ((c) >= 'a' && (c) <= 'z' ? (c) - 'a' + 'A' : (c))

/* Case insensitive string compare */
BOOL StringCompare(LPHUGE s1,LPHUGE s2,UINT Length1,UINT Length2)
{
    register WORD i;

    if(Length1 != Length2) return FALSE;

    i = (LPSTR)(MAKELP (SELECTOROF(s1), 0xFFFF)) - (LPSTR) s1 + 1 ;
    if(Length1 <= i)
    {
    	// Only one piece.
    	AnsiUpperBuff((LPSTR)s1,Length1);
    }
    else
    {
    	// String spans to segments.
    	AnsiUpperBuff((LPSTR)s1,i);
    	AnsiUpperBuff((LPSTR)(s1+i),Length1-i);
    }

    i = (LPSTR)(MAKELP (SELECTOROF(s2), 0xFFFF)) - (LPSTR) s2 + 1 ;
    if(Length2 <= i)
    {
    	// Only one piece.
    	AnsiUpperBuff((LPSTR)s2,Length2);
    }
    else
    {
    	// String spans to segments.
    	AnsiUpperBuff((LPSTR)s2,i);
    	AnsiUpperBuff((LPSTR)(s2+i),Length2-i);
    }

    for(i = 0; i < Length1; i++) if(s1[i] != s2[i]) return FALSE;
    return TRUE;
}

WORD LookupSymbol(DWORD Index, WORD Length)
{
    register WORD i;

    for(i = 2; i < nSymbol; i++)    // Loop starts at 2, because entry 0 is the endmarker
                                    // and entry 1 is the error token
	{
        if(StringCompare(hpSymbol[i].hpSymbolStart,hpInputBuffer+Index,
           (UINT)hpSymbol[i].SymbolSize,(UINT)Length) == TRUE) return i;
	}

    return 0;
}

/* Lookup a symbol in the symbol table. If not found, create new entry and store symbol. */
BOOL InstallSymbol(WORD SymbolType, DWORD Index, WORD Length, LPWORD Entry)
{
    *Entry = LookupSymbol(Index,Length);
    if(*Entry)
    {
        if(SymbolType == SYMBOL_TOKEN)
        {
            WORD i;

            LexErrorCode = PERR_TOKEN_MULT_DECLARED;
            for(i = 0; i < Length && i < MAX_ERROR_PARAM; i++)
                LexErrorStringParam[i] = hpInputBuffer[Index + i];  // Copy token to error parameter
            LexErrorStringParam[i] = 0; // Truncate to MAX_ERROR_PARAM
            return FALSE;
        }
        else return TRUE;
    }

    /* Symbol not found: create new entry */
    if(AddSymbol() == FALSE) return FALSE;
    hpSymbol[nSymbol - 1].TermUsed = FALSE;
    hpSymbol[nSymbol - 1].hpSymbolStart = hpInputBuffer + Index;
    hpSymbol[nSymbol - 1].SymbolSize = Length;
    hpSymbol[nSymbol - 1].nTerm = 0;
    hpSymbol[nSymbol - 1].nDesc = 0;
    hpSymbol[nSymbol - 1].nLeftProd = 0;
    *Entry = nSymbol - 1;
    return TRUE;
}

/* Analyze the source parser file, and store data in intermediate structures */
BOOL ParserPrepare()
{
    static DWORD Index,Index2;
    static WORD iSymbol;
    BOOL AfterWhiteSpace,ActionFound;
    LPHUGE p;
    unsigned char c;
    WORD LeftSymbol,ProdLength,Assoc;
    short int Precedence;

    Index = Index2 = 0;
    if(AddSymbol() == FALSE) return FALSE;      // Create the ENDMARKER in the Symbol array
    hpSymbol[0].TermUsed = TRUE;
    hpSymbol[0].hpSymbolStart = (LPHUGE)"$";
    hpSymbol[0].SymbolSize = 1;
    hpSymbol[0].nTerm = 0;
    hpSymbol[0].nDesc = 0;
    hpSymbol[0].nLeftProd = 0;

    if(AddSymbol() == FALSE) return FALSE;      // Create the ERROR_TOKEN in the Symbol array
    hpSymbol[1].TermUsed = TRUE;
    hpSymbol[1].hpSymbolStart = (LPHUGE)"error";
    hpSymbol[1].SymbolSize = 5;
    hpSymbol[1].nTerm = 0;
    hpSymbol[1].nDesc = 0;
    hpSymbol[1].nLeftProd = 0;

    nTerm=2;

    Precedence = 0;

    /* Part 1: Token declarations: %token ...
       It is mandatory to declare all tokens.
       The code fragment is copied to the produced file, without the enclosing { }. */
    while(1)
    {
        AfterWhiteSpace = FALSE;
        switch(hpInputBuffer[Index])
        {
          case 0:   LexErrorCode = PERR_NO_RULE_IN_BUFFER; return FALSE;
          case '%': Index++; yycol++;
                    p = hpInputBuffer + Index;
                    switch(p[0])
                    {
                        case '%':
                            Index++; yycol++; goto PART2;
                        case 'l':
                        case 'L':
                        case 'r':
                        case 'R':
                        case 't':
                        case 'T':
                            if(up(p[1]) == 'O' && up(p[2]) == 'K' && up(p[3]) == 'E' && up(p[4]) == 'N'
                               && IsCharAlpha(p[5]) == FALSE && p[5] != '_')
                            {
                                /* %token line: list of token declarations, separated by commas and whitespace,
                                   to the end of the line */
                                yycol += 4;
                                Index += 4;
                                Assoc = ASSOC_NONE;
                            }
                            else
                            {
                                if(up(p[1]) == 'E' && up(p[2]) == 'F' && up(p[3]) == 'T'
                                    && IsCharAlpha(p[4]) == FALSE && p[4] != '_')
                                {
                                    /* %left line: list of token declarations, separated by commas and whitespace,
                                       to the end of the line. These are declared to be left associative, and of
                                       same precedence */
                                    yycol += 3;
                                    Index += 3;
                                    Assoc = ASSOC_LEFT;
                                    Precedence++;
                                }
                                else
                                {
                                    if(up(p[1]) == 'I' && up(p[2]) == 'G' && up(p[3]) == 'H' && up(p[4]) == 'T'
                                       && IsCharAlpha(p[5]) == FALSE && p[5] != '_')
                                    {
                                        /* %right line: list of token declarations, separated by commas and whitespace,
                                           to the end of the line. These are declared to be right associative, and of
                                           same precedence */
                                        yycol += 4;
                                        Index += 4;
                                        Assoc = ASSOC_RIGHT;
                                        Precedence++;
                                    }
                                    else {LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = p[0]; return FALSE;}
                                }
                            }

                            LexErrorLine = yyline; LexErrorColumn = yycol;
                            do
                            {
                                yycol++; Index++;
                                if(ReadWhiteSpace(hpInputBuffer + Index, FALSE, &Index2) == FALSE) return FALSE;
                                Index += Index2;
                                if(hpInputBuffer[Index] == 13 || hpInputBuffer[Index] == 10 || hpInputBuffer[Index] == 0)
                                {LexErrorCode = PERR_INCOMPLETE_TOKEN; return FALSE;}
                                if(ReadId(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;
                                if(InstallSymbol(
                                    SYMBOL_TOKEN,
                                    Index,
                                    (WORD)Index2,
                                    &iSymbol) == FALSE) return FALSE;       // Register the token

                                /* Write token precedence and associativity */
                                if(Assoc == ASSOC_NONE)
                                    hpSymbol[nSymbol - 1].Precedence = 0;
                                else
                                    hpSymbol[nSymbol - 1].Precedence = (Assoc == ASSOC_LEFT ? Precedence : - Precedence);

                                nTerm++;                // Increment terminals count
                                Index += Index2;
                                
                                /* Skip whitespace */
                                if(ReadWhiteSpace(hpInputBuffer + Index, FALSE, &Index2) == FALSE) return FALSE;
                                Index += Index2;
                                
                                c = hpInputBuffer[Index];
                                if(c != 13 && c != 10 && c != ',')
                                {LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = c; return FALSE;}
                            } while(c == ',');
                            break;
                        default:
                            LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = p[0];
                            return FALSE;
                    }
                    break;
          case '{': P1Code.hpCodeStart = hpInputBuffer + Index;
                    P1Code.LineNumber = yyline;
                    if(ReadCode(hpInputBuffer + Index,&Index2) == FALSE) return FALSE;
                    P1Code.CodeSize = Index2 + 1;
                    Index += Index2;
                    break;
          case ' ':
          case 13:
          case '\t':
          case 10:  if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                    Index += Index2 - 1;
                    AfterWhiteSpace = TRUE;
                    break;
        
          default:
                    if(hpInputBuffer[Index] == '/' && (hpInputBuffer[Index+1] == '*' || hpInputBuffer[Index+1] == '/'))
                    {
                        /* A Comment ! */
                        if(ReadComment(hpInputBuffer+Index,&Index2) == FALSE) return FALSE;
                        Index += Index2;
                        break;
                    }
                    else    // Error
                    {
                        LexErrorCode = LEXERR_INVALID_CHARACTER;
                        LexErrorCharParam = hpInputBuffer[Index];
                        return FALSE;
                    }
        }
        if(AfterWhiteSpace == FALSE) yycol++;
        Index++;
    }

    /* Part 2: List of productions. Each production may have a semantic action. This action will
       later by copied each into the case of a switch statement, and the enclosing { } will be removed. */
    PART2:
    while(1)
    {
        AfterWhiteSpace = FALSE;
        switch(hpInputBuffer[Index])
        {
            case 0:     if(nProduction == 0) {LexErrorCode = PERR_NO_RULE_IN_BUFFER; return FALSE;}
                        else return TRUE;   // No Part 3
            case ' ':
            case 13:
            case '\t':
            case 10:    if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE; // Skip Whitespace
                        Index += Index2 - 1;
                        AfterWhiteSpace = TRUE;
                        break;
            case '%':   Index++; yycol++;
                        if(hpInputBuffer[Index] == '%')
                        {
                            Index++;
                            yycol++;
                            if(nProduction == 0) {LexErrorCode = PERR_NO_RULE_IN_BUFFER; return FALSE;}
                            goto PART3;
                        }
                        else {LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = hpInputBuffer[Index]; return FALSE;}
        
            default:    /*** A production or a comment ***/

                        /* A comment ? */
                        if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                        Index += Index2;
        
                        /*** A production ! ***/
                        if(ReadId(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;   // Read left nonterminal

                        if(nProduction == 0)
                        {
                            /* It is the first production: create the artificial production S' -> S */
                            if(AddSymbol() == FALSE) return FALSE;      // Create the new start symbol S'
                            hpSymbol[nSymbol - 1].TermUsed = FALSE;
                            hpSymbol[nSymbol - 1].hpSymbolStart = (LPHUGE)"Phoenix_Start_Symbol";
                            hpSymbol[nSymbol - 1].nTerm = 0;
                            hpSymbol[nSymbol - 1].nDesc = 0;
                            hpSymbol[nSymbol - 1].nLeftProd = 0;
                            hpSymbol[nSymbol - 1].SymbolSize = lstrlen((LPSTR)hpSymbol[nSymbol - 1].hpSymbolStart);
                            if(AddProduction() == FALSE) return FALSE;
                            hpProduction[nProduction - 1].Left = nSymbol - 1;
                            hpProduction[nProduction - 1].Length = 1;
                        }

                        /* Lookup symbol table for that nonterminal. If not found, create new entry. */
                        if(InstallSymbol(SYMBOL_NONTERMINAL,Index,(WORD)Index2,(LPWORD)&iSymbol) == FALSE) return FALSE;
                        LeftSymbol = iSymbol;
                        if(nProduction == 1) hpProduction[nProduction - 1].Right[0] = LeftSymbol;   // Right side of S' -> S
                        if(AddProduction() == FALSE) return FALSE;

                        /* Register the production in the index giving the list of productions having the same left side */
                        if(hpSymbol[LeftSymbol].nLeftProd >= NLEFTPROD) {LexErrorCode = PERR_NLEFTPROD_TOO_LOW; return FALSE;}
                        hpSymbol[LeftSymbol].LeftProd[hpSymbol[LeftSymbol].nLeftProd++] = nProduction - 1;

                        hpProduction[nProduction - 1].Left = LeftSymbol;
                        hpProduction[nProduction - 1].Length = 0;
                        hpProduction[nProduction - 1].hpActionStart = NULL;
                        Index += Index2;

                        /* Skip whitespace and colon */
                        if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                        Index += Index2;
                        if(hpInputBuffer[Index] != ':') {LexErrorCode = PERR_MISSING_SEMICOLON; return FALSE;}
                        Index++; yycol++;
                        if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                        Index += Index2;

                        ActionFound = FALSE;

                        while(1)
                        {
                            p = hpInputBuffer + Index;
                            switch(p[0])
                            {
                                case 0:     // Error: reached EOF
                                    LexErrorCode = LEXERR_REACHED_EOF_PRODUCTION; return FALSE;

                                case '{':   // A piece of code (semantic action)
                                    hpProduction[nProduction - 1].ActionLineNumber = yyline;
                                    if(ReadCode(hpInputBuffer + Index,&Index2) == FALSE) return FALSE;
                                    hpProduction[nProduction - 1].hpActionStart = hpInputBuffer + Index;
                                    hpProduction[nProduction - 1].ActionSize = Index2 + 1;
                                    Index += Index2 + 1;
                                    ActionFound = TRUE;
                                    break;

                                case '|':   // A new production with same left nonterminal
                                    if(AddProduction() == FALSE) return FALSE;

                                    /* Register the production in the index giving the list of productions having same left side */
                                    if(hpSymbol[LeftSymbol].nLeftProd >= NLEFTPROD)
                                    {LexErrorCode = PERR_NLEFTPROD_TOO_LOW; return FALSE;}
                                    hpSymbol[LeftSymbol].LeftProd[hpSymbol[LeftSymbol].nLeftProd++] = nProduction - 1;

                                    hpProduction[nProduction - 1].Left = LeftSymbol;
                                    hpProduction[nProduction - 1].Length = 0;
                                    hpProduction[nProduction - 1].hpActionStart = NULL;
                                    Index++; yycol++;
                                    ActionFound = FALSE;
                                    break;

                                case '%':                   // May be %prec directive
                                    if(up(p[1]) == 'P' && up(p[2]) == 'R' && up(p[3]) == 'E' && up(p[4]) == 'C'
                                        && IsCharAlpha(p[5]) == FALSE && p[5] != '_')
                                    {
                                        yycol += 5;
                                        Index += 5;

                                        /*** %prec directive: write production precedence and associativity ***/
                                        if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                                        Index += Index2;
                                        if(ReadId(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;
                                        if((iSymbol = LookupSymbol(Index,(WORD)Index2)) == 0 || iSymbol >= nTerm)
                                        {LexErrorCode = PERR_UNDECLARED_TOKEN; return FALSE;}
                                        Index += Index2;
                                        hpProduction[nProduction - 1].Precedence = hpSymbol[iSymbol].Precedence;
                                    }
                                    else {LexErrorCode = LEXERR_INVALID_CHARACTER; LexErrorCharParam = p[0]; return FALSE;}
                                    break;

                                case ';': goto END_PROD;    // End of the production

                                default:    // A symbol or a comment
                                    if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                                    Index += Index2;

                                    /* Verify that the semantic action was not previously found in production. */
                                    if(ActionFound) {LexErrorCode = PERR_SYMBOL_AFTER_ACTION; return FALSE;}

                                    /* Look for a possible error virtual token (error production). If found,
                                       it must be the first symbol on the right side of the production. */
                                    ProdLength = hpProduction[nProduction - 1].Length;
                                    p = hpInputBuffer + Index;
                                    if(up(p[0]) == 'E' && up(p[1]) == 'R' && up(p[2]) == 'R' && up(p[3]) == 'O'
                                       && up(p[4]) == 'R' && IsCharAlpha(p[5]) == FALSE && p[5] != '_')
                                    {
                                        /* error virtual token */
                                        if(ProdLength > 0)
                                        {LexErrorCode = PERR_ERROR_TOKEN_MISPLACED; return FALSE;}
                                        else
                                        {
                                            iSymbol = ERROR_TOKEN;
                                            Index += 5;
                                            yycol += 5;
                                        }
                                    }
                                    else
                                    {
                                        /* Normal symbol: read it */
                                        if(ReadId(hpInputBuffer + Index, &Index2) == FALSE) return FALSE;

                                        /* Lookup symbol table for that nonterminal. If not found, create new entry. */
                                        if(InstallSymbol(SYMBOL_UNDETERMINED,Index,(WORD)Index2,(LPWORD)&iSymbol) == FALSE) return FALSE;

                                        /* If symbol is a terminal, copy precedence to production (the precedence of a
                                           production is the precedence of it's righmost terminal).
                                           Additionally, set the "used" flag to TRUE. */
                                        if(iSymbol < nTerm)
                                        {
                                            hpProduction[nProduction - 1].Precedence = hpSymbol[iSymbol].Precedence;
                                            hpSymbol[iSymbol].TermUsed = TRUE;
                                        }

                                        Index += Index2;
                                    }

                                    if(ProdLength == LENPROD - 1) {LexErrorCode = PERR_TOO_LONG_PRODUCTION; return FALSE;}
                                    hpProduction[nProduction - 1].Right[ProdLength] = iSymbol;
                                    hpProduction[nProduction - 1].Length++;
                                    break;
                            }

                            /* Skip whitespace */
                            if(ReadWhiteSpace(hpInputBuffer + Index, TRUE, &Index2) == FALSE) return FALSE;
                            Index += Index2;
                        }
                        END_PROD:
                        break;
        }
        if(AfterWhiteSpace == FALSE) yycol++;
        Index++;
    }

    /* Part 3: support routines and code. Need not be enclosed in { }. */
    PART3:
    if(hpInputBuffer[Index] == 0) return TRUE;  // No code in Part 3
    P3Code.hpCodeStart = hpInputBuffer + Index;
    P3Code.LineNumber = yyline;
    Index2 = Index;
    while(hpInputBuffer[Index2]) Index2++;
    P3Code.CodeSize = Index2 - Index;

    return TRUE;
}
