// lexdoc.cpp : implementation of the CLexDoc class
//

#include "stdafx.h"
#include "phoenix.h"
#include "mainfrm.h"
#include "lexdoc.h"
#include "ph_start.h"
#include "packdlg.h"
#include "targtsys.h"
#include "extcompi.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CLexDoc

IMPLEMENT_DYNCREATE(CLexDoc, CDocument)

BEGIN_MESSAGE_MAP(CLexDoc, CDocument)
    //{{AFX_MSG_MAP(CLexDoc)
    ON_COMMAND(ID_COMPILE_GEN, OnCompileGen)
    ON_COMMAND(ID_OPTIONS_PACKING, OnOptionsPacking)
    ON_COMMAND(ID_OPTIONS_TARGETSYSTEM, OnOptionsTargetSys)
    ON_COMMAND(ID_OPTIONS_COLOR, OnOptionsColor)
    ON_COMMAND(ID_COMPILE_EXTERNALFILE, OnCompileExternalFile)
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

COLORREF CLexDoc::m_color = RGB(0,0,0);   // Lexical file text is black by default

/////////////////////////////////////////////////////////////////////////////
// CLexDoc construction/destruction

CLexDoc::CLexDoc()
{
    m_bPacked = TRUE;       // Arrays Packed by default
    m_nPF = DEFAULT_PF;     // Default Packing Factor
    m_nTarget = TARGET_UNIX;
}

CLexDoc::~CLexDoc()
{
}

BOOL CLexDoc::OnNewDocument()
{
    if (!CDocument::OnNewDocument())
        return FALSE;

    return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
// CLexDoc serialization

void CLexDoc::Serialize(CArchive& ar)
{
    // Serialize raw text contained in the first view
    POSITION pos = GetFirstViewPosition();
    if(pos != NULL)
    {
        CView *pView = GetNextView(pos);
        pView -> Serialize(ar);
    }
}


/////////////////////////////////////////////////////////////////////////////
// CLexDoc diagnostics

#ifdef _DEBUG
void CLexDoc::AssertValid() const
{
    CDocument::AssertValid();
}

void CLexDoc::Dump(CDumpContext& dc) const
{
    CDocument::Dump(dc);
}

#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CLexDoc commands

void CLexDoc::OnCompileGen()
{
    // If document needs to be saved, inform the user
    if(!SaveModified())
        return;

    CWnd *pWnd = AfxGetApp() -> m_pMainWnd; // Get main window
    if(IsModified())
    {
        pWnd -> MessageBox("Please save your lexical file before compiling it.",AfxGetAppName());
        return;
    }

    HWND hWnd = pWnd -> m_hWnd; // Get main window's handle

    // Load the Phoenix DLL
    HINSTANCE hPhoenixDll = LoadLibrary("PHX.DLL");
    if(hPhoenixDll < HINSTANCE_ERROR)
    {
        MessageBeep((UINT)-1);
        MessageBox(hWnd, "Failed to load PHX.DLL",AfxGetAppName(),
                   MB_APPLMODAL|MB_OK);
    }

    const char *FileName = GetPathName();
    char Path[128];
    LPSTR lpF;
    lstrcpy((LPSTR)Path,(LPSTR)FileName);
    lpF = (LPSTR)Path + lstrlen(Path);
    while(*lpF != '\\') lpF --;
    *lpF ='\0'; // Strip out the file name, keep the path
    HCURSOR hCur = SetCursor(LoadCursor(NULL,IDC_WAIT));

    // Compile lex file
    BOOL result = PhoenixAnalyzeLexFile(hWnd,(LPSTR)FileName,(LPSTR)Path,m_nTarget,
        NO_LINE_DIRECTIVES,(m_bPacked ? PACKED_ARRAYS : UNPACKED_ARRAYS),m_nPF);
    SetCursor(hCur);

    MessageBeep((UINT)-1);

    if(result == FALSE)
        MessageBox(hWnd,"Failed to generate lexical analyzer !",AfxGetAppName(),MB_OK);
    else
        MessageBox(hWnd,"Lexical analyzer produced !",AfxGetAppName(),MB_OK);

    FreeLibrary(hPhoenixDll);
}

void CLexDoc::OnOptionsPacking()
{
    // Invoking the Packing dialog box
    CPackDlg dlg;
    dlg.m_bPacked = m_bPacked;
    dlg.m_nPF = m_nPF;
    if(dlg.DoModal() == IDOK)
    {
        m_bPacked = dlg.m_bPacked;
        m_nPF = dlg.m_nPF;
    }
}

void CLexDoc::OnOptionsTargetSys()
{
    // Invoking the Target System dialog box
    CTargetSys dlg;
    dlg.m_nTarget = m_nTarget;
    if(dlg.DoModal() == IDOK)
    {
        m_nTarget = dlg.m_nTarget;
    }
}

void CLexDoc::OnOptionsColor()
{
    // Changing the color of all views of lexical type
    CColorDialog dlgColor(m_color);
    if(dlgColor.DoModal() == IDOK)
    {
        m_color = dlgColor.GetColor();
        ((CMainFrame *)AfxGetApp() -> m_pMainWnd) -> UpdateTextColor();
    }
}

void CLexDoc::OnCompileExternalFile()
{
    CExtCompile dlg;
    dlg.m_ofn.lpTemplateName = MAKEINTRESOURCE(IDD_EXTCOMPILE);
    dlg.m_ofn.hInstance = AfxGetInstanceHandle();
    dlg.m_ofn.lpstrFile = "";
    dlg.DoModal();
}
