// parsedoc.cpp : implementation file
//

#include "stdafx.h"
#include "phoenix.h"
#include "mainfrm.h"
#include "parsedoc.h"
#include "ph_start.h"
#include "targtsys.h"
#include "extcompi.h"

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CParseDoc

IMPLEMENT_DYNCREATE(CParseDoc, CDocument)

COLORREF CParseDoc::m_color = RGB(0,0,255); // Parser file text is blue by default

CParseDoc::CParseDoc()
{
    m_bPacked = TRUE;       // Arrays Packed by default
    m_nTarget = TARGET_UNIX;
}

BOOL CParseDoc::OnNewDocument()
{
    if (!CDocument::OnNewDocument())
        return FALSE;
    return TRUE;
}

CParseDoc::~CParseDoc()
{
}


BEGIN_MESSAGE_MAP(CParseDoc, CDocument)
    //{{AFX_MSG_MAP(CParseDoc)
    ON_COMMAND(ID_COMPILE_GEN, OnCompileGen)
    ON_COMMAND(ID_OPTIONS_PACKARR, OnOptionsPackarr)
    ON_UPDATE_COMMAND_UI(ID_OPTIONS_PACKARR, OnUpdateOptionsPackarr)
    ON_COMMAND(ID_OPTIONS_TARGETSYSTEM, OnOptionsTargetSys)
    ON_COMMAND(ID_OPTIONS_COLOR, OnOptionsColor)
    ON_COMMAND(ID_COMPILE_EXTERNALFILE, OnCompileExternalFile)
    //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CParseDoc serialization

void CParseDoc::Serialize(CArchive& ar)
{
    // Serialize raw text contained in the first view
    POSITION pos = GetFirstViewPosition();
    if(pos != NULL)
    {
        CView *pView = GetNextView(pos);
        pView -> Serialize(ar);
    }
}

/////////////////////////////////////////////////////////////////////////////
// CParseDoc commands

void CParseDoc::OnCompileGen()
{
    // If document needs to be saved, inform the user
    if(!SaveModified())
        return;

    CWnd *pWnd = AfxGetApp() -> m_pMainWnd; // Get main window
    if(IsModified())
    {
        pWnd -> MessageBox("Please save your parser file before compiling it.",AfxGetAppName());
        return;
    }

    HWND hWnd = pWnd -> m_hWnd; // Get main window's handle

    // Load the Phoenix DLL
    HINSTANCE hPhoenixDll = LoadLibrary("PHX.DLL");
    if(hPhoenixDll < HINSTANCE_ERROR)
    {
        MessageBeep((UINT)-1);
        MessageBox(hWnd, "Failed to load PHX.DLL",AfxGetAppName(),
                   MB_APPLMODAL|MB_OK);
    }

    const char *FileName = GetPathName();
    char Path[128];
    LPSTR lpF;
    lstrcpy((LPSTR)Path,(LPSTR)FileName);
    lpF = (LPSTR)Path + lstrlen(Path);
    while(*lpF != '\\') lpF --;
    *lpF ='\0'; // Strip out the file name, keep the path
    HCURSOR hCur = SetCursor(LoadCursor(NULL,IDC_WAIT));

    // Compile parser file
    BOOL result = PhoenixAnalyzeParserFile(hWnd,(LPSTR)FileName,(LPSTR)Path,m_nTarget,
        NO_LINE_DIRECTIVES,m_bPacked);
    SetCursor(hCur);

    MessageBeep((UINT)-1);

    if(result == FALSE)
        MessageBox(hWnd,"Failed to generate parser !",AfxGetAppName(),MB_OK);
    else
        MessageBox(hWnd,"Parser produced !",AfxGetAppName(),MB_OK);

    FreeLibrary(hPhoenixDll);
}

void CParseDoc::OnOptionsPackarr()
{
    m_bPacked = m_bPacked == TRUE ? FALSE : TRUE; // Change pack state
}

void CParseDoc::OnUpdateOptionsPackarr(CCmdUI* pCmdUI)
{
    pCmdUI -> SetCheck(m_bPacked);
}

void CParseDoc::OnOptionsTargetSys()
{
    // Invoking the Target System dialog box
    CTargetSys dlg;
    dlg.m_nTarget = m_nTarget;
    if(dlg.DoModal() == IDOK)
    {
        m_nTarget = dlg.m_nTarget;
    }
}

void CParseDoc::OnOptionsColor()
{
    // Changing the color of all views of parser type
    CColorDialog dlgColor(m_color);
    if(dlgColor.DoModal() == IDOK)
    {
        m_color = dlgColor.GetColor();
        ((CMainFrame *)AfxGetApp() -> m_pMainWnd) -> UpdateTextColor();
    }
}

void CParseDoc::OnCompileExternalFile()
{
    CExtCompile dlg;
    dlg.m_ofn.lpTemplateName = MAKEINTRESOURCE(IDD_EXTCOMPILE);
    dlg.m_ofn.hInstance = AfxGetInstanceHandle();
    dlg.m_ofn.lpstrFile = "";
    dlg.DoModal();
}
